/* Copyright (c) 2004-2018. The SimGrid Team. All rights reserved.          */

/* This program is free software; you can redistribute it and/or modify it
 * under the terms of the license (GNU LGPL) which comes with this package. */

#ifndef SIMGRID_KERNEL_RESOURCE_ACTION_HPP
#define SIMGRID_KERNEL_RESOURCE_ACTION_HPP

#include <simgrid/forward.h>
#include <xbt/signal.hpp>
#include <xbt/utility.hpp>

#include <boost/heap/pairing_heap.hpp>
#include <boost/optional.hpp>

const int NO_MAX_DURATION = -1.0;

namespace simgrid {
namespace kernel {
namespace resource {

typedef std::pair<double, simgrid::kernel::resource::Action*> heap_element_type;
typedef boost::heap::pairing_heap<heap_element_type, boost::heap::constant_time_size<false>, boost::heap::stable<true>,
                                  boost::heap::compare<simgrid::xbt::HeapComparator<heap_element_type>>>
    heap_type;

/** @ingroup SURF_interface
 * @brief SURF action interface class
 * @details An action is an event generated by a resource (e.g.: a communication for the network)
 */
class XBT_PUBLIC Action {
public:
  boost::intrusive::list_member_hook<> modifiedSetHook_; /* Used by the lazy update to list the actions to track */
  bool isLinkedModifiedSet() const { return modifiedSetHook_.is_linked(); }

  typedef boost::intrusive::member_hook<Action, boost::intrusive::list_member_hook<>, &Action::modifiedSetHook_>
      ActionLmmOptions;
  typedef boost::intrusive::list<Action, ActionLmmOptions> ActionLmmList;

  boost::intrusive::list_member_hook<> stateSetHook_;
  typedef boost::intrusive::member_hook<Action, boost::intrusive::list_member_hook<>, &Action::stateSetHook_>
      ActionOptions;
  typedef boost::intrusive::list<Action, ActionOptions> ActionList;

  enum class State {
    ready = 0,        /**< Ready        */
    running,          /**< Running      */
    failed,           /**< Task Failure */
    done,             /**< Completed    */
    to_free,          /**< Action to free in next cleanup */
    not_in_the_system /**< Not in the system anymore. Why did you ask ? */
  };

  enum class SuspendStates {
    not_suspended = 0, /**< Action currently not suspended **/
    suspended,
    sleeping
  };

  enum class Type { LATENCY = 100, MAX_DURATION, NORMAL, NOTSET };

  /**
   * @brief Action constructor
   *
   * @param model The Model associated to this Action
   * @param cost The cost of the Action
   * @param failed If the action is impossible (e.g.: execute something on a switched off host)
   */
  Action(simgrid::kernel::resource::Model* model, double cost, bool failed);

  /**
   * @brief Action constructor
   *
   * @param model The Model associated to this Action
   * @param cost The cost of the Action
   * @param failed If the action is impossible (e.g.: execute something on a switched off host)
   * @param var The lmm variable associated to this Action if it is part of a LMM component
   */
  Action(simgrid::kernel::resource::Model* model, double cost, bool failed, kernel::lmm::Variable* var);

  virtual ~Action();

  /**
   * @brief Mark that the action is now finished
   *
   * @param state the new [state](\ref simgrid::kernel::resource::Action::State) of the current Action
   */
  void finish(Action::State state);

  /** @brief Get the [state](\ref simgrid::kernel::resource::Action::State) of the current Action */
  Action::State getState() const; /**< get the state*/
  /** @brief Set the [state](\ref simgrid::kernel::resource::Action::State) of the current Action */
  virtual void setState(Action::State state);

  /** @brief Get the bound of the current Action */
  double getBound() const;
  /** @brief Set the bound of the current Action */
  void setBound(double bound);

  /** @brief Get the start time of the current action */
  double getStartTime() const { return start_; }
  /** @brief Get the finish time of the current action */
  double getFinishTime() const { return finishTime_; }

  /** @brief Get the user data associated to the current action */
  void* getData() const { return data_; }
  /** @brief Set the user data associated to the current action */
  void setData(void* data) { data_ = data; }

  /** @brief Get the cost of the current action */
  double getCost() const { return cost_; }
  /** @brief Set the cost of the current action */
  void setCost(double cost) { cost_ = cost; }

  /** @brief Update the maximum duration of the current action
   *  @param delta Amount to remove from the MaxDuration */
  void updateMaxDuration(double delta);

  /** @brief Update the remaining time of the current action
   *  @param delta Amount to remove from the remaining time */
  void updateRemains(double delta);

  /** @brief Set the remaining time of the current action */
  void setRemains(double value) { remains_ = value; }
  /** @brief Get the remaining time of the current action after updating the resource */
  virtual double getRemains();
  /** @brief Get the remaining time of the current action without updating the resource */
  double getRemainsNoUpdate() const { return remains_; }

  /** @brief Set the finish time of the current action */
  void setFinishTime(double value) { finishTime_ = value; }

  /**@brief Add a reference to the current action (refcounting) */
  void ref();
  /** @brief Unref that action (and destroy it if refcount reaches 0)
   *  @return true if the action was destroyed and false if someone still has references on it
   */
  virtual int unref();

  /** @brief Cancel the current Action if running */
  virtual void cancel();

  /** @brief Suspend the current Action */
  virtual void suspend();

  /** @brief Resume the current Action */
  virtual void resume();

  /** @brief Returns true if the current action is running */
  virtual bool isSuspended();

  /** @brief Get the maximum duration of the current action */
  double getMaxDuration() const { return maxDuration_; }
  /** @brief Set the maximum duration of the current Action */
  virtual void setMaxDuration(double duration);

  /** @brief Get the tracing category associated to the current action */
  char* getCategory() const { return category_; }
  /** @brief Set the tracing category of the current Action */
  void setCategory(const char* category);

  /** @brief Get the priority of the current Action */
  double getPriority() const { return sharingWeight_; };
  /** @brief Set the priority of the current Action */
  virtual void setSharingWeight(double priority);
  void setSharingWeightNoUpdate(double weight) { sharingWeight_ = weight; }

  /** @brief Get the state set in which the action is */
  ActionList* getStateSet() const { return stateSet_; };

  simgrid::kernel::resource::Model* getModel() const { return model_; }

protected:
  ActionList* stateSet_;
  int refcount_ = 1;

private:
  double sharingWeight_ = 1.0;             /**< priority (1.0 by default) */
  double maxDuration_   = NO_MAX_DURATION; /*< max_duration (may fluctuate until the task is completed) */
  double remains_;           /**< How much of that cost remains to be done in the currently running task */
  double start_;             /**< start time  */
  char* category_ = nullptr; /**< tracing category for categorized resource utilization monitoring */
  double finishTime_ =
      -1; /**< finish time : this is modified during the run and fluctuates until the task is completed */

  double cost_;
  simgrid::kernel::resource::Model* model_;
  void* data_ = nullptr; /**< for your convenience */

  /* LMM */
  double lastUpdate_                                  = 0;
  double lastValue_                                   = 0;
  kernel::lmm::Variable* variable_                    = nullptr;
  Action::Type type_                                  = Action::Type::NOTSET;
  boost::optional<heap_type::handle_type> heapHandle_ = boost::none;

public:
  virtual void updateRemainingLazy(double now) = 0;
  void heapInsert(heap_type& heap, double key, Action::Type hat);
  void heapRemove(heap_type& heap);
  void heapUpdate(heap_type& heap, double key, Action::Type hat);
  void clearHeapHandle() { heapHandle_ = boost::none; }
  kernel::lmm::Variable* getVariable() const { return variable_; }
  void setVariable(kernel::lmm::Variable* var) { variable_ = var; }
  double getLastUpdate() const { return lastUpdate_; }
  void refreshLastUpdate();
  double getLastValue() const { return lastValue_; }
  void setLastValue(double val) { lastValue_ = val; }
  Action::Type getType() const { return type_; }

protected:
  Action::SuspendStates suspended_ = Action::SuspendStates::not_suspended;
};

typedef Action::ActionList ActionList;
typedef Action::ActionLmmList ActionLmmList;
typedef Action::ActionLmmList* ActionLmmListPtr;
} // namespace resource
} // namespace kernel
} // namespace simgrid
#endif
